; ------------------------------------------------------
; OpenGL framework
; Copyright (c) 2005, Franck Charlet
; All rights reserved.
;
; Redistribution and use in source and binary forms, with or without modification,
; are permitted provided that the following conditions are met:
;
; * Redistributions of source code must retain the above copyright notice,
;   this list of conditions and the following disclaimer.
; * Redistributions in binary form must reproduce the above copyright notice,
;   this list of conditions and the following disclaimer in the documentation
;   and/or other materials provided with the distribution.
; * Neither the name of the <ORGANIZATION> nor the names of it's contributors
;   may be used to endorse or promote products derived from this software
;   without specific prior written permission.
;
; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
; AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
; THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
; IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
; INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
; (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
; HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
; OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
; ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
; ------------------------------------------------------

			.386
			.model	flat,stdcall
			option	casemap:none

; ------------------------------------------------------
; Includes
			include windows.inc
			include kernel32.inc
			include user32.inc
			include gdi32.inc
			include Opengl32.inc

			includelib kernel32.lib
			includelib user32.lib
			includelib gdi32.lib
			includelib opengl32.lib

; ------------------------------------------------------
; Macros
literal_float		MACRO	lit_float:VARARG
			LOCAL	local_float
			.data
local_float		real4	lit_float
			.code
			EXITM	<local_float>
			ENDM

CFLT			MACRO	lit_float:VARARG
			EXITM	<literal_float(lit_float)>
			ENDM

; ------------------------------------------------------
; structures
Application		struct
hInstance		dword	?
Application		ends

GL_WindowInit		struct
application		Application <>
winwidth		dword	?
winheight		dword	?
bitsPerPixel		dword	?
GL_WindowInit		ends

GL_Window		struct
hWnd			dword	?
hDC			dword	?
hRC			dword	?
init			GL_WindowInit <>
isVisible		dword	?
LastTickCount		LONGLONG ?
FirstTickCount		LONGLONG ?
GL_Window		ends

; ------------------------------------------------------
; Variables
			.data
			
g_window		GL_Window <>
g_keys			dd	0
g_ClassName		db	"OGLW",0
g_AskFullScreen		db	"Run in fullscreen ?",0
g_FullScreen		dd	0
g_CursorState		dd	0
_SubEditBrush		LOGBRUSH <BS_SOLID, 0, 0>
_wtop			dd	0
_wleft			dd	0
_Flt0001		real4	0.001
_fH			real8	0.0049999998882413
_zNear			real8	0.01
_zFar			real8	500.0
pfd			PIXELFORMATDESCRIPTOR <	sizeof PIXELFORMATDESCRIPTOR, \
						1, \
						04h or 020h or 1, \
						0, \
						0, \
						0, 0, 0, 0, 0, 0, \
						0, \
						0, \
						0, \
						0, 0, 0, 0, \
						16, \
						0, \
						0, \
						0, \
						0, \
						0, 0, 0 >
HTimerFreq		LONGLONG 0

; ------------------------------------------------------
; Public functions
Terminate_Application	proto
Initialize		proto
Deinitialize		proto
Update			proto	MilliSeconds:real4, TotalTime:real4
Draw			proto

			include OglUser.asm

; ------------------------------------------------------
; Name: Terminate_Application()
; Desc: Initiate termination procedure
Terminate_Application	proc
			invoke	PostMessage, dword ptr [g_window + GL_Window.hWnd], WM_QUIT, 0, 0
			ret
Terminate_Application	endp

; ------------------------------------------------------
; Name: Change_Screen_Resolution()
; Desc: Resize the screen with given dimensions and depth
Change_Screen_Resolution proc	w:dword, h:dword, bitsPerPixel:dword
			local	dmScreenSettings:DEVMODE

			invoke	RtlZeroMemory,addr dmScreenSettings, sizeof DEVMODE
			mov	dmScreenSettings.dmSize, sizeof DEVMODE
			mov	eax,w
			mov	dmScreenSettings.dmPelsWidth,eax
			mov	eax,h
			mov	dmScreenSettings.dmPelsHeight,eax
			mov	eax,bitsPerPixel
			mov	dmScreenSettings.dmBitsPerPel,eax
			mov	dmScreenSettings.dmFields, 040000h or 080000h or 0100000h
			invoke	ChangeDisplaySettings,addr dmScreenSettings, 4
			.if	eax != DISP_CHANGE_SUCCESSFUL
				xor	eax,eax
			.else
				mov	eax,TRUE
			.endif
			ret
Change_Screen_Resolution endp

; -----------------------------------------------------------------------
; Name: WindowProc()
; Desc: Messages handler for the main window
WindowProc		proc	hWnd:dword, uMsg:UINT, wParam:WPARAM, lParam:LPARAM
			local	creation:CREATESTRUCT
			
			.if	uMsg == WM_SYSCOMMAND
				.if	wParam == SC_SCREENSAVE || wParam == SC_MONITORPOWER
					xor	eax,eax
					ret
				.endif
			.elseif	uMsg == WM_CREATE
				xor	eax,eax
				ret
			.elseif	uMsg == WM_CLOSE
				invoke	Terminate_Application
				xor	eax,eax
				ret
			.elseif	uMsg == WM_ACTIVATE
				.if	g_FullScreen != 0
					mov	eax,wParam
					and	eax,0ffffh
					mov	ecx,eax
					and	eax,WA_ACTIVE
					and	ecx,WA_CLICKACTIVE
					.if	eax != 0 || ecx != 0
						invoke	Change_Screen_Resolution, dword ptr [g_window + GL_Window.init.winwidth], dword ptr [g_window + GL_Window.init.winheight], dword ptr [g_window + GL_Window.init.bitsPerPixel]
						invoke	GetSystemMetrics, SM_CXSCREEN
						push	eax
						invoke	GetSystemMetrics, SM_CYSCREEN
						mov	ecx,eax
						pop	eax
						invoke	SetWindowPos, hWnd, HWND_TOPMOST, 0, 0, eax, ecx, SWP_SHOWWINDOW
						mov	dword ptr [g_window + GL_Window.isVisible],TRUE
					.else
						invoke	ChangeDisplaySettings, NULL, 0
						invoke	SetWindowPos, hWnd, HWND_BOTTOM, 0, 0, 0, 0, SWP_SHOWWINDOW
						invoke	ShowWindow,hWnd, SW_MINIMIZE
						mov	dword ptr [g_window + GL_Window.isVisible],FALSE
					.endif
				.endif
			.elseif	uMsg == WM_SIZE
				.if	wParam == SIZE_MINIMIZED
					mov	dword ptr [g_window + GL_Window.isVisible],FALSE
				.elseif wParam == SIZE_MAXIMIZED
					mov	dword ptr [g_window + GL_Window.isVisible],TRUE
				.elseif wParam == SIZE_RESTORED
					mov	dword ptr [g_window + GL_Window.isVisible],TRUE
				.endif
			.elseif	uMsg == WM_KEYDOWN
				.if wParam >= 0 && wParam <= 255
					mov	eax,[g_keys]
					mov	ecx,wParam
					mov	byte ptr [eax + ecx], TRUE
					xor	eax,eax
					ret
				.endif
			.elseif	uMsg == WM_MOUSEMOVE
				.if	g_CursorState == TRUE
					mov	g_CursorState, FALSE
					invoke	ShowCursor, FALSE
				.endif
			.elseif	uMsg == WM_NCMOUSEMOVE
				.if g_CursorState == FALSE
					mov	g_CursorState, TRUE
					invoke	ShowCursor, TRUE
				.endif
			.elseif	uMsg == WM_KEYUP
				.if wParam >= 0 && wParam <= 255
					mov	eax,[g_keys]
					mov	ecx,wParam
					mov	byte ptr [eax + ecx], FALSE
					xor	eax,eax
					ret
				.endif
			.endif
			invoke	DefWindowProc, hWnd, uMsg, wParam, lParam
			ret
WindowProc		endp

; -----------------------------------------------------------------------
; Name: Main()
; Desc: Program entry point
Main			proc
			local	msg:MSG
			local	LastT:real4
			local	FirstT:real4
			local	windowClass:WNDCLASSEX
			local	windowStyle:dword
			local	windowExtendedStyle:dword
			local	wwidth:dword
			local	wheight:dword
			local	PixelFormat:dword
			local	_fW:real8
			local	_mfW:real8
			local	_mfH:real8
			local	HTimerVal:LARGE_INTEGER

			lea	edi,g_window
			lea	ebx,windowClass
			invoke	RtlZeroMemory, ebx, sizeof WNDCLASSEX
			invoke	GetModuleHandle,NULL
			mov	[ebx + WNDCLASSEX.hInstance],eax
			push	eax
			pop	dword ptr [edi + GL_Window.init.application.hInstance]
			mov	dword ptr [edi + GL_Window.init.winwidth], SCREEN_WIDTH
			mov	dword ptr [edi + GL_Window.init.winheight], SCREEN_HEIGHT
			mov	dword ptr [edi + GL_Window.init.bitsPerPixel], SCREEN_BPP
			mov	dword ptr [ebx + WNDCLASSEX.cbSize], sizeof WNDCLASSEX
			mov	dword ptr [ebx + WNDCLASSEX.style], CS_HREDRAW or CS_VREDRAW or CS_OWNDC
			push	offset WindowProc
			pop	dword ptr [ebx + WNDCLASSEX.lpfnWndProc]
			invoke	CreateBrushIndirect, addr _SubEditBrush
			mov	[ebx + WNDCLASSEX.hbrBackground],eax
			push	offset g_ClassName
			pop	dword ptr [ebx + WNDCLASSEX.lpszClassName]
			invoke	RegisterClassEx, addr windowClass
			.if	eax == 0
				ret
			.endif
			
			invoke	MessageBox, 0, addr g_AskFullScreen, offset g_WindowName, MB_YESNOCANCEL or MB_ICONQUESTION
			.if	eax == IDYES
				mov	g_FullScreen, TRUE
			.elseif	eax == IDNO
			.elseif	eax == IDCANCEL
				ret
			.endif
			
; Force window mode if constant _DEBUG is defined
ifdef _DEBUG
			mov	g_FullScreen, FALSE
endif
			.if	g_FullScreen == TRUE
				mov	windowStyle, WS_POPUP
				mov	windowExtendedStyle, WS_EX_TOPMOST
			.else
				mov	windowStyle, WS_BORDER or WS_CAPTION or WS_SYSMENU or WS_POPUP or WS_MINIMIZEBOX
				mov	windowExtendedStyle, WS_EX_APPWINDOW
			.endif

			mov	eax,dword ptr [g_window + GL_Window.init.bitsPerPixel]
			mov	[pfd.cColorBits],al

			mov	eax,dword ptr [g_window + GL_Window.init.winwidth]
			mov	[wwidth],eax
			mov	eax,dword ptr [g_window + GL_Window.init.winheight]
			mov	[wheight],eax
			invoke	GetSystemMetrics, SM_CXSCREEN
			.if	wwidth > eax
				mov	wwidth,eax
			.endif
			invoke	GetSystemMetrics, SM_CYSCREEN
			.if	wheight > eax
				mov	wheight,eax
			.endif
			
			.if	g_FullScreen == 0
				; Center the window if necessary
				invoke	GetSystemMetrics, SM_CXSCREEN
				.if	wwidth < eax
					sub	eax,wwidth
					shr	eax,1
					mov	_wleft,eax
				.endif

				invoke	GetSystemMetrics, SM_CYSCREEN
				.if	wheight < eax
					sub	eax,wheight
					shr	eax,1
					mov	_wtop,eax
				.endif
				; Add some pixels around to make the inner part to fit the dimensions requirements
				invoke	GetSystemMetrics, SM_CXBORDER
				add	wwidth,eax
				invoke	GetSystemMetrics, SM_CYBORDER
				add	wheight,eax
				invoke	GetSystemMetrics, SM_CYCAPTION
				add	wheight,eax
			.endif

			or	windowStyle,WS_VISIBLE
			invoke	CreateWindowEx, windowExtendedStyle, \
							offset g_ClassName, \
							offset g_WindowName, \
							windowStyle, \
							_wleft, _wtop, \
							wwidth, \
							wheight, \
							HWND_DESKTOP, \
							0, \
							dword ptr [g_window + GL_Window.init.application.hInstance], \
							addr g_window
			mov	dword ptr [g_window + GL_Window.hWnd],eax
			.if	eax == 0
				ret
			.endif
			invoke	PeekMessage, addr msg, 0, 0, 0, PM_REMOVE
			.if	eax != 0
				invoke	TranslateMessage, addr msg
				invoke	DispatchMessage, addr msg
			.endif
			invoke	ShowCursor, FALSE
			.if	g_FullScreen != 0
				invoke	Change_Screen_Resolution, dword ptr [g_window + GL_Window.init.winwidth], dword ptr [g_window + GL_Window.init.winheight], dword ptr [g_window + GL_Window.init.bitsPerPixel]
				.if	eax == 0
					ret
				.endif
				invoke	MoveWindow, dword ptr [g_window + GL_Window.hWnd], 0, 0, dword ptr [g_window + GL_Window.init.winwidth], dword ptr [g_window + GL_Window.init.winheight], 1
			.endif
			invoke	GetDC, dword ptr [g_window + GL_Window.hWnd]
			mov	dword ptr [g_window + GL_Window.hDC],eax
			.if	eax == 0
				jmp	Remove_Window
			.endif
			invoke	ChoosePixelFormat, dword ptr [g_window + GL_Window.hDC], addr pfd
			mov	PixelFormat,eax
			.if	eax == 0
				jmp	Release_DC
			.endif
			invoke	SetPixelFormat, dword ptr [g_window + GL_Window.hDC], [PixelFormat], addr pfd
			.if eax == FALSE
				jmp	Release_DC
			.endif
			invoke	wglCreateContext, dword ptr [g_window + GL_Window.hDC]
			mov	dword ptr [g_window + GL_Window.hRC],eax 
			.if	eax == 0
				jmp	Release_DC
			.endif
			invoke	wglMakeCurrent, dword ptr [g_window + GL_Window.hDC], dword ptr [g_window + GL_Window.hRC]
			.if	eax == FALSE
				invoke	wglDeleteContext, dword ptr [g_window + GL_Window.hRC]
Release_DC:			invoke	ReleaseDC, dword ptr [g_window + GL_Window.hWnd], dword ptr [g_window + GL_Window.hDC]
Remove_Window:			invoke	DestroyWindow, dword ptr [g_window + GL_Window.hWnd]
				ret
			.endif
			mov	dword ptr [g_window + GL_Window.isVisible], TRUE

			; Set the projection matrix
			invoke	glViewport, 0, 0, dword ptr [g_window + GL_Window.init.winwidth], dword ptr [g_window + GL_Window.init.winheight]
			invoke	glMatrixMode, GL_PROJECTION
			invoke	glLoadIdentity
			fld	qword ptr [_fH]
			fld	st(0)
			fchs
			fstp	qword ptr [_mfH]
			fild	dword ptr [g_window + GL_Window.init.winwidth]
			fidiv	dword ptr [g_window + GL_Window.init.winheight]
			fmulp	st(1),st(0)
			fst	qword ptr [_fW]
			fchs
			fstp	qword ptr [_mfW]
			invoke	glFrustum, dword ptr [_mfW], dword ptr [_mfW + 4], dword ptr [_fW], dword ptr [_fW + 4], dword ptr [_mfH], dword ptr [_mfH + 4], dword ptr [_fH], dword ptr [_fH + 4], dword ptr [_zNear], dword ptr [_zNear + 4], dword ptr [_zFar], dword ptr [_zFar + 4]
			invoke	glMatrixMode, GL_MODELVIEW
			invoke	glLoadIdentity

			invoke	LocalAlloc, LMEM_FIXED or LMEM_ZEROINIT, 256
			mov	g_keys,eax

			invoke	Initialize
			.if	eax == FALSE
				invoke	Terminate_Application
			.else
				; init counter
				invoke	QueryPerformanceFrequency, addr HTimerVal
				invoke	GetLastError
				fild	qword ptr [HTimerVal.QuadPart]
				fmul	dword ptr [_Flt0001]
				fistp	qword ptr [HTimerFreq]
				invoke	QueryPerformanceCounter, addr HTimerVal
				fild	qword ptr [HTimerVal.QuadPart]
				fistp	qword ptr [g_window + GL_Window.LastTickCount]
				fild	qword ptr [HTimerVal.QuadPart]
				fistp	qword ptr [g_window + GL_Window.FirstTickCount]
				mov	msg.message, 0
				.while	msg.message != WM_QUIT
					invoke	PeekMessage, addr msg, dword ptr [g_window + GL_Window.hWnd], 0, 0, PM_REMOVE
					.if	eax != 0
						invoke	DispatchMessage, addr msg
					.else
						.if	dword ptr [g_window + GL_Window.isVisible] == FALSE
							invoke	WaitMessage
						.else
							invoke	QueryPerformanceCounter, addr HTimerVal

							fild	qword ptr [HTimerVal.QuadPart]
							fild	qword ptr [g_window + GL_Window.LastTickCount]
							fsubp	st(1),st(0)
							fild	qword ptr [HTimerFreq]
							fdivp	st(1),st(0)
							fmul	dword ptr [_Flt0001]
							fstp	dword ptr [LastT]

							fild	qword ptr [HTimerVal.QuadPart]
							fild	qword ptr [g_window + GL_Window.FirstTickCount]
							fsubp	st(1),st(0)
							fild	qword ptr [HTimerFreq]
							fdivp	st(1),st(0)
							fmul	dword ptr [_Flt0001]
							fstp	dword ptr [FirstT]
							
							invoke	Update, LastT, FirstT
							fild	qword ptr [HTimerVal.QuadPart]
							fistp	qword ptr [g_window + GL_Window.LastTickCount]
							invoke	Draw
							invoke	SwapBuffers, dword ptr [g_window + GL_Window.hDC]
						.endif
					.endif
				.endw
			.endif
			invoke	Deinitialize
			invoke	ChangeDisplaySettings, NULL, 0
			.if	dword ptr [g_window + GL_Window.hWnd] != 0
				.if	dword ptr [g_window + GL_Window.hDC] != 0
					invoke	wglMakeCurrent, dword ptr [g_window + GL_Window.hDC], 0
					.if	dword ptr [g_window + GL_Window.hRC] != 0
						invoke	wglDeleteContext, dword ptr [g_window + GL_Window.hRC]
					.endif
					invoke	ReleaseDC, dword ptr [g_window + GL_Window.hWnd], dword ptr [g_window + GL_Window.hDC]
				.endif
				invoke	DestroyWindow, dword ptr [g_window + GL_Window.hWnd]
			.endif
			invoke	ShowCursor, TRUE
			invoke	UnregisterClass, offset g_ClassName, dword ptr [g_window + GL_Window.init.application.hInstance]
			ret
Main			endp

start:			invoke	Main
			invoke	ExitProcess,0

end start